from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import WaterfallStep
from weaverbird.pipeline.steps.waterfall import GROUP_WATERFALL_COLUMN

_VQB_ALL_KEYS = "_vqbAllKeys"
_VQB_START_KEYS = "_vqbStartKeys"
_VQB_END_KEYS = "_vqbEndKeys"
_VQB_START_ELEMENTS = "_vqbStartElements"
_VQB_END_ELEMENTS = "_vqbEndElements"
_VQB_CHILDREN = "_vqbChildren"
_VQB_PARENTS = "_vqbParents"
_VQB_KEYS_TO_KEEP = "_vqbKeysToKeep"
_VQB_START = "_vqbStart"
_VQB_END = "_vqbEnd"
_VQB_RESULTS = "_vqbResults"
_VQB_ORDER = "_vqbOrder"


def _facet_keys_and_elements(step: WaterfallStep, *, group_key: list[str], project_key: list[str]) -> list[MongoStep]:
    """Facets the input document in order to get:

    * A list of all existing group keys
    * A list of all existing group keys for start and end elements
    * A list of all start elements
    * A list of all end elements
    """

    # fields to keep for start and end element keys
    start_end_key_projection_fields = {
        "_id": False,
        **{col: f"$_id.{col}" for col in group_key + [step.milestonesColumn]},
    }
    # fields to keep for start and end elements
    start_end_projection_fields = {
        # We store the group key in the ID for later filtering
        "_id": {col: f"$_id.{col}" for col in group_key},
        **{col: f"$_id.{col}" for col in project_key if col != "_id"},
    }

    return [
        {
            "$facet": {
                _VQB_ALL_KEYS: [
                    {"$match": {step.milestonesColumn: {"$in": [step.start, step.end]}}},
                    {"$group": {"_id": _column_map(group_key)}},
                    {"$project": {"_id": False, **{col: f"$_id.{col}" for col in group_key}}},
                ],
                _VQB_START_KEYS: [
                    {"$match": {step.milestonesColumn: step.start}},
                    {"$group": {"_id": _column_map(group_key)}},
                    {"$project": start_end_key_projection_fields},
                ],
                _VQB_START_ELEMENTS: [
                    {"$match": {step.milestonesColumn: step.start}},
                    {"$group": {"_id": _column_map(project_key)}},
                    {"$project": start_end_projection_fields},
                ],
                _VQB_END_KEYS: [
                    {"$match": {step.milestonesColumn: step.end}},
                    {"$group": {"_id": _column_map(group_key)}},
                    {"$project": start_end_key_projection_fields},
                ],
                _VQB_END_ELEMENTS: [
                    {"$match": {step.milestonesColumn: step.end}},
                    {"$group": {"_id": _column_map(project_key)}},
                    {"$project": start_end_projection_fields},
                ],
            }
        }
    ]


def _filter_out_incomplete_elements(*, group_key: list[str], project_key: list[str]) -> list[MongoStep]:
    """Filters out elements which do not have a start and end value"""
    return [
        # Determining which keys should be kept: We only what those with a start and end value
        {"$addFields": {_VQB_KEYS_TO_KEEP: {"$setIntersection": [f"${_VQB_START_KEYS}", f"${_VQB_END_KEYS}"]}}},
        # filtering start and elements on keys that should be kept
        {
            "$project": {
                _VQB_START_ELEMENTS: {
                    "$filter": {
                        "input": f"${_VQB_START_ELEMENTS}",
                        "cond": {"$in": ["$$this._id", f"${_VQB_KEYS_TO_KEEP}"]},
                    }
                },
                _VQB_END_ELEMENTS: {
                    "$filter": {
                        "input": f"${_VQB_END_ELEMENTS}",
                        "cond": {"$in": ["$$this._id", f"${_VQB_KEYS_TO_KEEP}"]},
                    }
                },
            }
        },
    ]


def _backfill_missing_values(step: WaterfallStep, *, group_key: list[str], project_key: list[str]) -> list[MongoStep]:
    """Backfills the missing start and end values."""
    mongo_step = {
        "$project": {
            _VQB_START_ELEMENTS: {
                # Concatenating the actual start keys + the backfilled ones
                "$concatArrays": [
                    f"${_VQB_START_ELEMENTS}",
                    # building the backfilled elements: mapping over the missing keys and adding the
                    # label + the value
                    {
                        "$map": {
                            # Determining the missing keys by doing a set difference between all
                            # keys and the start keys
                            "input": {"$setDifference": [f"${_VQB_ALL_KEYS}", f"${_VQB_START_KEYS}"]},
                            "in": {
                                "$mergeObjects": [
                                    "$$this",
                                    {step.milestonesColumn: step.start, step.valueColumn: 0},
                                ]
                            },
                        }
                    },
                ]
            },
            _VQB_END_ELEMENTS: {
                "$concatArrays": [
                    f"${_VQB_END_ELEMENTS}",
                    {
                        "$map": {
                            "input": {"$setDifference": [f"${_VQB_ALL_KEYS}", f"${_VQB_END_KEYS}"]},
                            "in": {
                                "$mergeObjects": [
                                    "$$this",
                                    {step.milestonesColumn: step.start, step.valueColumn: 0},
                                ]
                            },
                        }
                    },
                ]
            },
        }
    }
    return [mongo_step]


def _calculate_children_deltas(step: WaterfallStep) -> list[MongoStep]:
    """Calculates the deltas between the start and the end of every element.

    The deltas are added to a new field.
    """
    mongo_step = {
        "$addFields": {
            _VQB_CHILDREN: {
                # Iterating over a zip op (end_element, start_element) pairs
                "$map": {
                    "input": {"$zip": {"inputs": [f"${_VQB_END_ELEMENTS}", f"${_VQB_START_ELEMENTS}"]}},
                    "in": {
                        # Here, we are merging the start object with another object containg the
                        # value column with the start element's value subtracted from the end
                        # element's value
                        "$mergeObjects": [
                            {"$arrayElemAt": ["$$this", 0]},
                            {
                                step.valueColumn: {
                                    "$subtract": [
                                        {
                                            "$getField": {
                                                "field": step.valueColumn,
                                                "input": {"$arrayElemAt": ["$$this", 0]},
                                            }
                                        },
                                        {
                                            "$getField": {
                                                "field": step.valueColumn,
                                                "input": {"$arrayElemAt": ["$$this", 1]},
                                            }
                                        },
                                    ]
                                }
                            },
                        ]
                    },
                }
            }
        }
    }
    return [mongo_step]


def _sort_elements(*, group_key: list[str], project_key: list[str]) -> list[MongoStep]:
    # FIXME: this should be $sortArray once we can use mongo>=5.2
    return [
        {
            "$facet": {
                _VQB_START_ELEMENTS: [
                    {"$unwind": {"path": f"${_VQB_START_ELEMENTS}"}},
                    {"$sort": {f"{_VQB_START_ELEMENTS}.{col}": 1 for col in group_key}},
                    {"$project": {col: f"${_VQB_START_ELEMENTS}.{col}" for col in project_key}},
                ],
                _VQB_END_ELEMENTS: [
                    {"$unwind": {"path": f"${_VQB_END_ELEMENTS}"}},
                    {"$sort": {f"{_VQB_END_ELEMENTS}.{col}": 1 for col in group_key}},
                    {"$project": {col: f"${_VQB_END_ELEMENTS}.{col}" for col in project_key}},
                ],
            }
        },
    ]


def _facet_results(step: WaterfallStep) -> list[MongoStep]:
    children_group = step.groupby + [step.labelsColumn] + ([step.parentsColumn] if step.parentsColumn else [])
    facet: dict[str, list] = {
        _VQB_CHILDREN: [
            {"$unwind": f"${_VQB_CHILDREN}"},
            {
                "$group": {
                    "_id": {col: f"${_VQB_CHILDREN}.{col}" for col in children_group},
                    step.valueColumn: {"$sum": f"${_VQB_CHILDREN}.{step.valueColumn}"},
                }
            },
            {
                "$project": {
                    **{col: f"$_id.{col}" for col in step.groupby},
                    "LABEL_waterfall": f"$_id.{step.labelsColumn}",
                    **({"GROUP_waterfall": f"$_id.{step.parentsColumn}"} if step.parentsColumn else {}),
                    "TYPE_waterfall": "child" if step.parentsColumn else "parent",
                    step.valueColumn: f"${step.valueColumn}",
                    _VQB_ORDER: {"$literal": 1},
                }
            },
        ],
        _VQB_START: [
            {"$unwind": f"${_VQB_START_ELEMENTS}"},
            {
                "$group": {
                    "_id": {col: f"${_VQB_START_ELEMENTS}.{col}" for col in step.groupby} if step.groupby else True,
                    step.valueColumn: {"$sum": f"${_VQB_START_ELEMENTS}.{step.valueColumn}"},
                }
            },
            {
                "$project": {
                    **{col: f"$_id.{col}" for col in step.groupby},
                    "LABEL_waterfall": {"$literal": str(step.start)},
                    "GROUP_waterfall": {"$literal": str(step.start)},
                    "TYPE_waterfall": None,
                    step.valueColumn: f"${step.valueColumn}",
                    _VQB_ORDER: {"$literal": 0},
                },
            },
        ],
        _VQB_END: [
            {"$unwind": f"${_VQB_END_ELEMENTS}"},
            {
                "$group": {
                    "_id": {col: f"${_VQB_END_ELEMENTS}.{col}" for col in step.groupby} if step.groupby else True,
                    step.valueColumn: {"$sum": f"${_VQB_END_ELEMENTS}.{step.valueColumn}"},
                }
            },
            {
                "$project": {
                    **{col: f"$_id.{col}" for col in step.groupby},
                    "LABEL_waterfall": {"$literal": str(step.end)},
                    "GROUP_waterfall": {"$literal": str(step.end)},
                    "TYPE_waterfall": None,
                    step.valueColumn: f"${step.valueColumn}",
                    _VQB_ORDER: {"$literal": 3},
                },
            },
        ],
    }

    if step.parentsColumn:
        facet[_VQB_PARENTS] = [
            {"$unwind": f"${_VQB_CHILDREN}"},
            {
                "$group": {
                    "_id": {col: f"${_VQB_CHILDREN}.{col}" for col in (step.groupby + [step.parentsColumn])},
                    step.valueColumn: {"$sum": f"${_VQB_CHILDREN}.{step.valueColumn}"},
                }
            },
            {
                "$project": {
                    **{col: f"$_id.{col}" for col in step.groupby},
                    "LABEL_waterfall": f"$_id.{step.parentsColumn}",
                    "GROUP_waterfall": f"$_id.{step.parentsColumn}",
                    "TYPE_waterfall": "parent",
                    step.valueColumn: f"${step.valueColumn}",
                    "_id": False,
                    _VQB_ORDER: {"$literal": 2},
                },
            },
        ]
    return [{"$facet": facet}]


def _unwind_results(step: WaterfallStep) -> list[MongoStep]:
    to_concat = [_VQB_START, _VQB_CHILDREN, _VQB_END]
    if step.parentsColumn:
        to_concat.append(_VQB_PARENTS)

    return [
        {"$project": {_VQB_RESULTS: {"$concatArrays": [f"${col}" for col in to_concat]}}},
        {"$unwind": f"${_VQB_RESULTS}"},
        {"$replaceRoot": {"newRoot": f"${_VQB_RESULTS}"}},
    ]


def _column_map(colnames: list[str]) -> dict[str, str]:
    return {col: f"${col}" for col in colnames}


def translate_waterfall(step: WaterfallStep) -> list[MongoStep]:
    group_key = step.groupby + ([step.parentsColumn] if step.parentsColumn else []) + [step.labelsColumn]
    project_key = group_key + [step.milestonesColumn, step.valueColumn, "_id"]

    steps = _facet_keys_and_elements(step, group_key=group_key, project_key=project_key)
    if step.backfill:
        # Backfilling missing values
        steps += _backfill_missing_values(step, group_key=group_key, project_key=project_key)
    else:
        # Filtering out imcomplete elements (where either start or end is missing)
        steps += _filter_out_incomplete_elements(group_key=group_key, project_key=project_key)

    # Sorting both arrays on their group key to have them in the same order
    steps += _sort_elements(group_key=group_key, project_key=project_key)
    # Calculating the delta between the end and the start of the period for every element
    steps += _calculate_children_deltas(step)
    # Facetting children, parents, start and end value
    steps += _facet_results(step)
    # Unwinding everything
    steps += _unwind_results(step)

    unset: str | list[str] = _VQB_ORDER
    if len(step.groupby) == 0 and not step.parentsColumn:
        unset = [_VQB_ORDER, GROUP_WATERFALL_COLUMN]

    return steps + [
        {
            "$sort": {
                _VQB_ORDER: 1,
                ("LABEL_waterfall" if step.sortBy == "label" else step.valueColumn): 1 if step.order == "asc" else -1,
            },
        },
        {"$unset": unset},
    ]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.cumsum import CumSumStep


def translate_cumsum(step: CumSumStep) -> list[MongoStep]:
    groupby = step.groupby or []

    return [
        {"$sort": {step.reference_column: 1}},
        {
            "$group": {
                "_id": {col: f"${col}" for col in groupby} if step.groupby else None,
                **{x: {"$push": f"${x}"} for x, _ in step.to_cumsum},
                "_vqbArray": {"$push": "$$ROOT"},
            },
        },
        {"$unwind": {"path": "$_vqbArray", "includeArrayIndex": "_VQB_INDEX"}},
        {
            "$project": {
                **{col: f"$_id.{col}" for col in groupby},
                **{
                    new_name if new_name else f"{name}_CUMSUM": {
                        "$sum": {"$slice": [f"${name}", {"$add": ["$_VQB_INDEX", 1]}]}
                    }
                    for name, new_name in step.to_cumsum
                },
                "_vqbArray": 1,
            },
        },
        {"$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbArray", "$$ROOT"]}}},
        {"$project": {"_vqbArray": 0}},
    ]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ToDateStep(BaseStep):
    name: Literal["todate"] = "todate"
    column: ColumnName
    format: str | None = None



import math


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"



from collections.abc import Sequence
from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.steps.utils.validation import validate_unique_columns
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

AggregateFn = Literal[
    "avg",
    "sum",
    "min",
    "max",
    "count",
    "count distinct",
    "first",
    "last",
    "count distinct including empty",
]


class Aggregation(BaseModel):
    new_columns: list[ColumnName] = Field(alias="newcolumns")
    agg_function: AggregateFn = Field(alias="aggfunction")
    columns: list[ColumnName]
    model_config = ConfigDict(populate_by_name=True)

    @field_validator("columns", mode="before")
    @classmethod
    def validate_unique_columns(cls, value):
        return validate_unique_columns(value)

    @model_validator(mode="before")
    @classmethod
    def handle_legacy_syntax(cls, values):
        if "column" in values:
            values["columns"] = [values.pop("column")]
        if "newcolumn" in values:
            values["new_columns"] = [values.pop("newcolumn")]
        return values


class AggregateStep(BaseStep):
    name: Literal["aggregate"] = "aggregate"
    on: list[ColumnName] = []
    aggregations: Sequence[Aggregation]
    keep_original_granularity: bool | None = False


class AggregationWithVariables(Aggregation):
    new_columns: list[TemplatedVariable] = Field(alias="newcolumns")
    agg_function: TemplatedVariable = Field(alias="aggfunction")
    columns: list[TemplatedVariable]


class AggregateStepWithVariables(AggregateStep, StepWithVariablesMixin):
    aggregations: Sequence[AggregationWithVariables]



from pypika import functions
from pypika.dialects import Query
from pypika.enums import Dialects
from pypika.queries import Selectable
from pypika.terms import Case, CustomFunction, Field, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateFormatMapping,
    SQLTranslator,
)


class ToMilliseconds(functions.Function):
    def __init__(self, field: Field):
        super().__init__("TO_MILLISECONDS", field)


class AthenaTranslator(SQLTranslator):
    DIALECT = SQLDialect.ATHENA
    QUERY_CLS = Query
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = False
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE",
        integer="INTEGER",
        text="VARCHAR",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://prestodb.io/docs/current/functions/datetime.html#mysql-date-functions
        day_number="%d",
        month_number="%m",
        month_short="%b",
        month_full="%M",
        year="%Y",
    )
    REGEXP_OP = RegexOp.REGEXP_LIKE
    TO_DATE_OP = ToDateOp.DATE_PARSE
    FROM_DATE_OP = FromDateOp.DATE_FORMAT

    @classmethod
    def _add_date(cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None) -> Term:
        # We need implement our own function for athena because Presto requires the units to be
        # quoted. PyPika's DateAdd function removes them by applying LiteralValue to the unit
        custom = CustomFunction("DATE_ADD", ["unit", "duration", "target"])
        return custom(Term.wrap_constant(unit.removesuffix("s")), duration, target_column)

    @staticmethod
    def _wrap_split_part(term: Term) -> Term:
        # Athena is the only backend for which SPLIT_PART returns NULL fox out-of-range
        # indexes. Also, IF_NULL is not available
        return Case().when(term.isnull(), "").else_(term)

    @classmethod
    def _interval_to_seconds(cls, value: Selectable) -> functions.Function:
        return ToMilliseconds(value) / 1000


SQLTranslator.register(AthenaTranslator)



from collections.abc import Sequence
from typing import Literal

from pydantic import Field, field_validator

from weaverbird.pipeline.steps.utils.base import BaseModel, BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

from .aggregate import Aggregation, AggregationWithVariables


class TotalDimension(BaseModel):
    total_column: ColumnName
    total_rows_label: str


class TotalsStep(BaseStep):
    name: Literal["totals"] = "totals"
    total_dimensions: list[TotalDimension]
    aggregations: Sequence[Aggregation]
    groups: list[ColumnName] = Field(min_length=0, default_factory=list)

    @field_validator("aggregations")
    @classmethod
    def aggregation_must_not_be_empty(cls, value):
        if len(value) < 1:
            raise ValueError("aggregations must contain at least one item")
        return value


class TotalsStepWithVariable(TotalsStep, StepWithVariablesMixin):
    aggregations: Sequence[AggregationWithVariables]
    groups: TemplatedVariable | list[TemplatedVariable]



from typing import Literal

from pydantic import Field, model_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class PercentageStep(BaseStep):
    name: Literal["percentage"] = "percentage"
    column: ColumnName
    group: list[ColumnName] = Field(default_factory=list)
    new_column_name: ColumnName | None = None

    @model_validator(mode="after")
    def set_new_column_name_if_not_set(self) -> "PercentageStep":
        if self.new_column_name is None:
            self.new_column_name = f"{self.column}_PCT"
        return self



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class ReplaceTextStep(BaseStep):
    name: Literal["replacetext"] = "replacetext"
    search_column: ColumnName
    old_str: str
    new_str: str = ""


class ReplaceTextStepWithVariable(ReplaceTextStep, StepWithVariablesMixin):
    old_str: TemplatedVariable



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import RankStep

_TMP_INDEX_COLUMN_NAME = "__VQB_TMP_INDEX_COLUMN__"


def execute_rank(
    step: RankStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column_name = step.new_column_name or f"{step.value_col}_RANK"
    rank_method = "min" if step.method == "standard" else step.method
    ascending = step.order == "asc"
    if step.groupby:
        serie = df.groupby(step.groupby, dropna=False)[step.value_col]
    else:
        serie = df[step.value_col]
    rank_serie = serie.rank(method=rank_method, ascending=ascending)
    return (
        df.assign(**{new_column_name: rank_serie})
        # NOTE: Sorting on a temporary index column as well in order to preserve the order
        .reset_index(names=[_TMP_INDEX_COLUMN_NAME])
        .sort_values(by=[new_column_name, _TMP_INDEX_COLUMN_NAME])
        .drop([_TMP_INDEX_COLUMN_NAME], axis=1)
    )



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import PercentageStep


def translate_percentage(step: PercentageStep) -> list[MongoStep]:
    return [
        {
            "$group": {
                "_id": {g: f"${g}" for g in step.group} if step.group else None,
                "_vqbAppArray": {"$push": "$$ROOT"},
                "_vqbTotalDenum": {"$sum": f"${step.column }"},
            },
        },
        {"$unwind": "$_vqbAppArray"},
        {
            "$project": {
                step.new_column_name: {
                    "$cond": [
                        {"$eq": ["$_vqbTotalDenum", 0]},
                        None,
                        {"$divide": [f"$_vqbAppArray.{step.column}", "$_vqbTotalDenum"]},
                    ],
                },
                "_vqbAppArray": 1,  # we need to keep track of this key for the next operation
            },
        },
        # Line below: Keep all columns that were not used in computation, 'stored' in _vqbAppArray
        {"$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbAppArray", "$$ROOT"]}}},
        {"$project": {"_vqbAppArray": 0}},  # We do not want to keep that column at the end
    ]



from collections.abc import Callable
from typing import Any

from weaverbird.backends.mongo_translator.steps.absolutevalue import translate_absolutevalue
from weaverbird.backends.mongo_translator.steps.addmissingdates import translate_addmissingdates
from weaverbird.backends.mongo_translator.steps.aggregate import translate_aggregate
from weaverbird.backends.mongo_translator.steps.append import translate_append
from weaverbird.backends.mongo_translator.steps.argmax import translate_argmax
from weaverbird.backends.mongo_translator.steps.argmin import translate_argmin
from weaverbird.backends.mongo_translator.steps.comparetext import translate_comparetext
from weaverbird.backends.mongo_translator.steps.concatenate import translate_concatenate
from weaverbird.backends.mongo_translator.steps.convert import translate_convert
from weaverbird.backends.mongo_translator.steps.cumsum import translate_cumsum
from weaverbird.backends.mongo_translator.steps.custom import translate_custom
from weaverbird.backends.mongo_translator.steps.date_extract import translate_date_extract
from weaverbird.backends.mongo_translator.steps.delete import translate_delete
from weaverbird.backends.mongo_translator.steps.domain import translate_domain
from weaverbird.backends.mongo_translator.steps.duplicate import translate_duplicate
from weaverbird.backends.mongo_translator.steps.duration import translate_duration
from weaverbird.backends.mongo_translator.steps.evolution import translate_evolution
from weaverbird.backends.mongo_translator.steps.fillna import translate_fillna
from weaverbird.backends.mongo_translator.steps.filter import translate_filter
from weaverbird.backends.mongo_translator.steps.formula import translate_formula
from weaverbird.backends.mongo_translator.steps.fromdate import translate_fromdate
from weaverbird.backends.mongo_translator.steps.ifthenelse import translate_ifthenelse
from weaverbird.backends.mongo_translator.steps.join import translate_join
from weaverbird.backends.mongo_translator.steps.lowercase import translate_lowercase
from weaverbird.backends.mongo_translator.steps.moving_average import translate_moving_average
from weaverbird.backends.mongo_translator.steps.percentage import translate_percentage
from weaverbird.backends.mongo_translator.steps.pivot import translate_pivot
from weaverbird.backends.mongo_translator.steps.rank import translate_rank
from weaverbird.backends.mongo_translator.steps.rename import translate_rename
from weaverbird.backends.mongo_translator.steps.replace import translate_replace
from weaverbird.backends.mongo_translator.steps.replacetext import translate_replacetext
from weaverbird.backends.mongo_translator.steps.rollup import translate_rollup
from weaverbird.backends.mongo_translator.steps.select import translate_select
from weaverbird.backends.mongo_translator.steps.sort import translate_sort
from weaverbird.backends.mongo_translator.steps.split import translate_split
from weaverbird.backends.mongo_translator.steps.statistics import translate_statistics
from weaverbird.backends.mongo_translator.steps.substring import translate_substring
from weaverbird.backends.mongo_translator.steps.text import translate_text
from weaverbird.backends.mongo_translator.steps.todate import translate_todate
from weaverbird.backends.mongo_translator.steps.top import translate_top
from weaverbird.backends.mongo_translator.steps.totals import translate_totals
from weaverbird.backends.mongo_translator.steps.trim import translate_trim
from weaverbird.backends.mongo_translator.steps.uniquegroups import translate_uniquegroups
from weaverbird.backends.mongo_translator.steps.unpivot import translate_unpivot
from weaverbird.backends.mongo_translator.steps.uppercase import translate_uppercase
from weaverbird.backends.mongo_translator.steps.waterfall import translate_waterfall

# I would like to have a better type for the callable, but I don't know how to do it.
# each of this function take a particular step as input
# so it is not possible to use `Dict[str, Callable[[BaseStep], list]]
mongo_step_translator: dict[str, Callable[[Any], list]] = {
    "absolutevalue": translate_absolutevalue,
    "addmissingdates": translate_addmissingdates,
    "aggregate": translate_aggregate,
    "append": translate_append,
    "argmax": translate_argmax,
    "argmin": translate_argmin,
    "comparetext": translate_comparetext,
    "concatenate": translate_concatenate,
    "convert": translate_convert,
    "cumsum": translate_cumsum,
    "custom": translate_custom,
    "dateextract": translate_date_extract,
    "delete": translate_delete,
    "domain": translate_domain,
    "duplicate": translate_duplicate,
    "duration": translate_duration,
    "evolution": translate_evolution,
    "fillna": translate_fillna,
    "filter": translate_filter,
    "formula": translate_formula,
    "fromdate": translate_fromdate,
    "ifthenelse": translate_ifthenelse,
    "join": translate_join,
    "lowercase": translate_lowercase,
    "movingaverage": translate_moving_average,
    "percentage": translate_percentage,
    "pivot": translate_pivot,
    "rename": translate_rename,
    "rank": translate_rank,
    "replace": translate_replace,
    "replacetext": translate_replacetext,
    "rollup": translate_rollup,
    "select": translate_select,
    "sort": translate_sort,
    "split": translate_split,
    "statistics": translate_statistics,
    "substring": translate_substring,
    "text": translate_text,
    "todate": translate_todate,
    "totals": translate_totals,
    "top": translate_top,
    "trim": translate_trim,
    "uniquegroups": translate_uniquegroups,
    "unpivot": translate_unpivot,
    "uppercase": translate_uppercase,
    "waterfall": translate_waterfall,
}



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.pivot import PivotStep


def translate_pivot(step: PivotStep) -> list[MongoStep]:
    group_cols = {}
    add_fields_step = {}

    # Prepare groupCols to populate the `_id` field sof Mongo `$group` steps and addFields step
    for col in step.index:
        group_cols[col] = f"$_id.{col}"
        add_fields_step[f"_vqbAppTmpObj.{col}"] = f"$_id.{col}"

    pivot_mongo_agg_stages: list[MongoStep] = [
        # First we perform the aggregation with the _id including the column to pivot
        {
            "$group": {
                "_id": {
                    **{col: f"${col}" for col in step.index},
                    step.column_to_pivot: f"${step.column_to_pivot}",
                },
                step.value_column: {f"${step.agg_function}": f"${step.value_column}"},
            },
        },
        # Then we group with with index columns as _id and we push documents as an array of sets
        # including a column for the column to pivot and a column for the corresponding value
        {
            "$group": {
                "_id": group_cols,
                "_vqbAppArray": {
                    "$addToSet": {
                        step.column_to_pivot: f"$_id.{step.column_to_pivot}",
                        step.value_column: f"${step.value_column}",
                    },
                },
            },
        },
        # Then we project a tmp key to get an object from the array of couples [column_to_pivot, corresponding_value]
        # including a column for the column to pivot and a column for the corresponding value
        {
            "$project": {
                "_vqbAppTmpObj": {
                    "$arrayToObject": {
                        "$zip": {
                            "inputs": [
                                f"$_vqbAppArray.{step.column_to_pivot}",
                                f"$_vqbAppArray.{step.value_column}",
                            ],
                        },
                    },
                },
            },
        },
    ]

    if add_fields_step:
        # Then we include the index columns back in every document created in the previous step
        # (still accessible in the _id object)
        pivot_mongo_agg_stages.append({"$addFields": add_fields_step})

    # Then we replace the root of the documents tree to get our columns ready for
    # our needed table-like, unnested format
    pivot_mongo_agg_stages.append({"$replaceRoot": {"newRoot": "$_vqbAppTmpObj"}})

    return pivot_mongo_agg_stages



from functools import reduce
from typing import Any

import numpy
import pandas as pd
from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.waterfall import (
    GROUP_WATERFALL_COLUMN,
    LABEL_WATERFALL_COLUMN,
    RESULT_COLUMN,
    TYPE_WATERFALL_COLUMN,
    WaterfallStep,
)

_VQB_SORT_ORDER = "_vqbSortOrder"
_VQB_GROUP = "_vqbGroup"

# Steps taken to execute a waterfall step:
# * Filter the input dataset on "label IN [start, end]"
# * If backfill is true, set 0 for all missing start/end values. If it is false, remove all rows for
#   groups which do not have a start/end value
# * For all groups ([label, (parent)] + groups), compute the difference between start and end labels
# * If there is a parent column, calculate the sums for the parent values as well
# * Return a dataframe in the following format
#     - One row with the total on start label
#     - One row with the delta for every group (label + groupby columns + optional parent)
#     - If there are parents, one row with the sum of deltas for all groups with this parent (can be
#       mixed up with children rows)
#     - One row with the total on end label


def execute_waterfall(
    step: WaterfallStep,
    df: DataFrame,
    domain_retriever: DomainRetriever | None = None,
    execute_pipeline: PipelineExecutor | None = None,
) -> DataFrame:
    # Start of the waterfall
    start_df = df[df[step.milestonesColumn] == step.start]
    # End of the waterfall
    end_df = df[df[step.milestonesColumn] == step.end]

    group_columns = _group_by_columns(step)
    start_value_groups = _unique_values_groups(start_df, group_columns)
    end_value_groups = _unique_values_groups(end_df, group_columns)
    unique_values_groups = start_value_groups.union(end_value_groups)

    # Backfilling missing values
    if step.backfill:
        start_df = _backfill_missing_values(start_df, step, step.labelsColumn, group_columns, unique_values_groups)
        end_df = _backfill_missing_values(end_df, step, step.labelsColumn, group_columns, unique_values_groups)
    # Otherwise, filter out rows which do not have a start and end date
    else:
        # We want to remove all value groups which are not in both the start and end dataframes
        value_groups_to_remove = unique_values_groups - start_value_groups.intersection(end_value_groups)
        start_df = _filter_out_rows(start_df, group_columns, value_groups_to_remove)
        end_df = _filter_out_rows(end_df, group_columns, value_groups_to_remove)

    start_df = start_df.rename(columns={step.valueColumn: f"{step.valueColumn}_start"})
    end_df = end_df.rename(columns={step.valueColumn: f"{step.valueColumn}_end"})

    merged = _merge(step, start_df, end_df)
    start_values = _compute_agg_milestone(
        step,
        df,
        start_df.rename(columns={f"{step.valueColumn}_start": step.valueColumn}),
        step.start,
        0,
    )
    parents_children = _compute_parents_and_children(step, merged)
    end_values = _compute_agg_milestone(
        step, df, end_df.rename(columns={f"{step.valueColumn}_end": step.valueColumn}), step.end, 3
    )
    result = pd.concat([start_values, parents_children, end_values])
    result.sort_values(
        by=[_VQB_SORT_ORDER, _get_sort_column(step)],
        ascending=[True, step.order == "asc"],
        inplace=True,
    )
    del result[_VQB_SORT_ORDER]
    return result


def _group_by_columns(step: WaterfallStep) -> list[str]:
    group_by_columns = [step.labelsColumn] + step.groupby
    if step.parentsColumn is not None:
        group_by_columns = group_by_columns + [step.parentsColumn]
    return group_by_columns


def _merge(step: WaterfallStep, start_df: DataFrame, end_df: DataFrame) -> DataFrame:
    """Computes the difference for the value of every label between end and start.

    :param start_df: Base DataFrame filtered to contain only results at the start of the waterfall
    :param end_df: Base DataFrame filtered to contain only results at the end of the waterfall
    """
    group_by_columns = _group_by_columns(step)
    start_df = (
        start_df.groupby(by=group_by_columns)
        .agg({step.valueColumn + "_start": "sum"})
        .merge(start_df[group_by_columns], on=group_by_columns)
        .drop_duplicates()
    )

    end_df = (
        end_df.groupby(by=group_by_columns)
        .agg({step.valueColumn + "_end": "sum"})
        .merge(end_df[group_by_columns], on=group_by_columns)
        .drop_duplicates()
    )

    # we join the result to compare them
    merged_df = start_df.merge(end_df, on=_get_join_key(step))
    merged_df[RESULT_COLUMN] = merged_df[f"{step.valueColumn}_end"] - merged_df[f"{step.valueColumn}_start"]
    merged_df = merged_df.drop(
        columns=[
            f"{step.valueColumn}_start",
            f"{step.valueColumn}_end",
        ]
    )

    # if there is a parent column, we need to aggregate for them
    if step.parentsColumn is not None:
        parents_results = merged_df.groupby(step.groupby + [step.parentsColumn], as_index=False).agg(
            {RESULT_COLUMN: "sum"}
        )
        parents_results[step.labelsColumn] = parents_results[step.parentsColumn]
        parents_results[TYPE_WATERFALL_COLUMN] = "parent"
        merged_df[TYPE_WATERFALL_COLUMN] = "child"
        return pd.concat([merged_df, parents_results])
    else:
        merged_df[TYPE_WATERFALL_COLUMN] = "parent"
    return merged_df


def _unique_values_groups(df: DataFrame, group_columns: list[str]) -> set[tuple[str, ...]]:
    return set(df[group_columns].drop_duplicates().to_records(index=False).tolist())


def _filter_out_rows(
    df: DataFrame, group_columns: list[str], value_groups_to_remove: set[tuple[str, ...]]
) -> DataFrame:
    if not value_groups_to_remove:
        return df

    for value_group_to_remove in value_groups_to_remove:
        conditions = (df[col] == val for col, val in zip(group_columns, value_group_to_remove, strict=True))
        condition = reduce(lambda s1, s2: s1 & s2, conditions)
        df = df[~condition]

    return df


def _backfill_missing_values(
    df: DataFrame,
    step: WaterfallStep,
    label_value: str,
    group_columns: list[str],
    unique_values_groups: set[tuple[str, ...]],
    backfill_value: int = 0,
) -> DataFrame:
    missing_value_groups = unique_values_groups - _unique_values_groups(df, group_columns)
    if not missing_value_groups:
        return df

    new_rows: list[dict[str, Any]] = []
    for missing_groups in missing_value_groups:
        new_rows.append(
            {
                step.labelsColumn: label_value,
                step.valueColumn: backfill_value,
                **dict(zip(group_columns, missing_groups, strict=True)),
            }
        )

    return pd.concat([df, pd.DataFrame(new_rows)])


def _compute_agg_milestone(
    step: WaterfallStep, df: DataFrame, start_df: DataFrame, milestone, sort_order: int
) -> pd.DataFrame:
    """Creates the top and bottom rows for waterfalls"""
    if len(step.groupby) > 0:
        groups = df[step.groupby].drop_duplicates()
        group_by = step.groupby
    else:
        groups = pd.DataFrame({_VQB_GROUP: [0]})  # pseudo group
        start_df[_VQB_GROUP] = 0
        groups[_VQB_GROUP] = 0
        group_by = [_VQB_GROUP]

    groups = groups.assign(**{step.labelsColumn: milestone})

    agg = start_df.groupby(group_by).agg({f"{step.valueColumn}": "sum"})
    agg = groups.merge(agg, on=group_by).rename(columns={step.labelsColumn: LABEL_WATERFALL_COLUMN})

    if step.parentsColumn is not None:
        agg[GROUP_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
    agg[TYPE_WATERFALL_COLUMN] = None
    agg[LABEL_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
    if len(step.groupby) == 0:
        del start_df[_VQB_GROUP]
        del agg[_VQB_GROUP]
    agg[_VQB_SORT_ORDER] = sort_order
    return agg


def _compute_parents_and_children(step: WaterfallStep, merged_df: DataFrame) -> DataFrame:
    """Computes children values (the deepest nesting level)"""
    result_df = DataFrame({LABEL_WATERFALL_COLUMN: merged_df[step.labelsColumn].astype(str)})
    result_df[step.groupby] = merged_df[step.groupby]
    result_df[TYPE_WATERFALL_COLUMN] = merged_df[TYPE_WATERFALL_COLUMN]

    if step.parentsColumn is None:
        result_df[_VQB_SORT_ORDER] = 2
    else:
        result_df[GROUP_WATERFALL_COLUMN] = merged_df[step.parentsColumn]
        result_df[_VQB_SORT_ORDER] = numpy.where(result_df[TYPE_WATERFALL_COLUMN] == "parent", 2, 1)

    result_df[step.valueColumn] = merged_df[RESULT_COLUMN]
    return result_df


def _get_sort_column(step: WaterfallStep):
    if step.sortBy == "value":
        return step.valueColumn
    else:
        return LABEL_WATERFALL_COLUMN


def _get_join_key(step: WaterfallStep):
    if step.parentsColumn is None:
        return [step.labelsColumn] + step.groupby
    else:
        return [step.labelsColumn, step.parentsColumn] + step.groupby



from collections.abc import Callable
from typing import Any

from .base import BaseStep


class StepWithVariablesMixin:
    def render(self, variables: dict[str, Any], renderer: Callable[[Any, Any], Any]) -> BaseStep:
        rendered_class = self.__class__.__bases__[0]
        step_as_dict = self.model_dump()  # type: ignore
        rendered_dict = renderer(step_as_dict, variables)
        return rendered_class(**rendered_dict)



from enum import Enum


class SQLDialect(str, Enum):
    # Note: please use the same translator names than in weaverbird's frontend
    ATHENA = "athena"
    GOOGLEBIGQUERY = "google-big-query"
    MYSQL = "mysql"
    POSTGRES = "postgresql"
    REDSHIFT = "redshift"
    SNOWFLAKE = "snowflake"



from weaverbird.backends.mongo_translator.steps.formula import build_mongo_formula_tree
from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.backends.mongo_translator.utils import build_cond_expression
from weaverbird.pipeline.formula_ast.eval import FormulaParser
from weaverbird.pipeline.steps.ifthenelse import IfThenElse, IfthenelseStep


def transform_ifthenelse_step(step: IfThenElse) -> MongoStep:
    else_expr: dict | str | int | float | bool
    if isinstance(step.else_value, IfThenElse):
        else_expr = transform_ifthenelse_step(step.else_value)
    else:
        try:
            else_expr = build_mongo_formula_tree(FormulaParser(str(step.else_value)).parse())
        except SyntaxError:  # else_expr is a badly formatted string
            else_expr = step.else_value

    if_expr = build_cond_expression(step.condition)
    try:
        then_expr = build_mongo_formula_tree(FormulaParser(str(step.then)).parse())
    except SyntaxError:  # step is a badly formatted string
        return step.then

    return {"$cond": {"if": if_expr, "then": then_expr, "else": else_expr}}


def translate_ifthenelse(step: IfthenelseStep) -> list[MongoStep]:
    return [
        {
            "$addFields": {
                step.new_column: transform_ifthenelse_step(
                    IfThenElse(condition=step.condition, then=step.then, else_value=step.else_value)
                )
            }
        }
    ]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ToDateStep


def translate_todate(step: ToDateStep) -> list[MongoStep]:
    date_format = step.format

    col_as_string = {"$toString": f"${step.column}"}

    if date_format:
        # % B and %b should be equivalent
        date_format = date_format.replace("%B", "%b")

    if not date_format:
        # Mongo will try to guess the date format
        return [
            {
                "$addFields": {
                    step.column: {
                        "$dateFromString": {
                            "dateString": col_as_string,
                            # However, we give it a second chance to find formats it can't guess natively
                            "onError": {
                                "$cond": [
                                    # Integer values may be either years or timestamps
                                    {"$in": [{"$type": f"${step.column}"}, ["int", "long"]]},
                                    {
                                        "$cond": [
                                            # We decide that values lower than 10_000 will be interpreted as years...
                                            {"$lt": [f"${step.column}", 10_000]},
                                            {
                                                "$dateFromString": {
                                                    "dateString": {"$concat": [col_as_string, "-01-01"]},
                                                    "format": "%Y-%m-%d",
                                                    "onError": {"$literal": None},
                                                }
                                            },
                                            # ...and greater values will be interpreted as timestamps
                                            {
                                                "$convert": {
                                                    "input": f"${step.column}",
                                                    "to": "date",
                                                    "onError": {"$literal": None},
                                                }
                                            },
                                        ]
                                    },
                                    # otherwise, try to interpret as years
                                    {
                                        "$dateFromString": {
                                            "dateString": {"$concat": [col_as_string, "-01-01"]},
                                            "format": "%Y-%m-%d",
                                            # and give up if it doesn't work
                                            "onError": {"$literal": None},
                                        }
                                    },
                                ]
                            },
                        }
                    }
                }
            }
        ]

    # Mongo doesn't handle months names (%b and %B), so we replace them by numbers before converting
    # All these could probably be factorized to make it easier to read!
    elif date_format == "%d %b %Y":
        return [
            {"$addFields": {"_vqbTempArray": {"$split": [col_as_string, " "]}}},
            _extract_date_parts_to_temp_fields(2, 1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column,
                [
                    "$_vqbTempDay",
                    "-",
                    "$_vqbTempMonth",
                    "-",
                    "$_vqbTempYear",
                ],
                "%d-%m-%Y",
            ),
            _clean_temp_fields(),
        ]

    elif date_format == "%d-%b-%Y":
        return [
            {"$addFields": {"_vqbTempArray": {"$split": [col_as_string, "-"]}}},
            _extract_date_parts_to_temp_fields(2, 1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column,
                [
                    "$_vqbTempDay",
                    "-",
                    "$_vqbTempMonth",
                    "-",
                    "$_vqbTempYear",
                ],
                "%d-%m-%Y",
            ),
            _clean_temp_fields(),
        ]

    elif date_format == "%b %Y":
        return [
            {"$addFields": {"_vqbTempArray": {"$split": [col_as_string, " "]}}},
            _extract_date_parts_to_temp_fields(1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(step.column, ["01-", "$_vqbTempMonth", "-", "$_vqbTempYear"], "%d-%m-%Y"),
            _clean_temp_fields(),
        ]

    elif date_format == "%b-%Y":
        return [
            {"$addFields": {"_vqbTempArray": {"$split": [col_as_string, "-"]}}},
            _extract_date_parts_to_temp_fields(1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(step.column, ["01-", "$_vqbTempMonth", "-", "$_vqbTempYear"], "%d-%m-%Y"),
            _clean_temp_fields(),
        ]

    # Mongo does not support dates where some parts of the date are missing
    elif date_format == "%Y-%m":
        return [
            _concat_fields_to_date(step.column, [col_as_string, "-01"], "%Y-%m-%d"),
        ]

    elif date_format == "%Y/%m":
        return [
            _concat_fields_to_date(step.column, ["/01", col_as_string], "%Y/%m/%d"),
        ]

    elif date_format == "%m-%Y":
        return [
            _concat_fields_to_date(step.column, ["01-", col_as_string], "%d-%m-%Y"),
        ]

    elif date_format == "%m/%Y":
        return [
            _concat_fields_to_date(step.column, ["01/", col_as_string], "%d/%m/%Y"),
        ]

    elif date_format == "%Y":
        return [
            _concat_fields_to_date(step.column, [col_as_string, "-01-01"], "%Y-%m-%d"),
        ]

    else:
        return [
            {
                "$addFields": {
                    step.column: {
                        "$dateFromString": {
                            "dateString": col_as_string,
                            "format": date_format,
                            "onError": {"$literal": None},
                        }
                    }
                }
            }
        ]


MONTH_NUMBER_TO_NAMES = {
    "01": ["jan", "jan.", "january", "janv", "janv.", "janvier"],
    "02": ["feb", "feb.", "february", "fév", "fev", "févr.", "fevr.", "février"],
    "03": ["mar", "mar.", "march", "mars"],
    "04": ["apr", "apr.", "april", "avr", "avr.", "avril"],
    "05": ["may", "mai"],
    "06": ["june", "jun.", "june", "juin"],
    "07": ["jul", "jul.", "july", "juil", "juil.", "juillet"],
    "08": ["aug", "aug.", "august", "août", "aout"],
    "09": ["sep", "sep.", "september", "sept", "sept.", "septembre"],
    "10": ["oct", "oct.", "october", "octobre"],
    "11": ["nov", "nov.", "november", "novembre"],
    "12": ["dec", "dec.", "december", "déc", "déc.", "décembre"],
}

MONTH_REPLACEMENT_STEP: MongoStep = {
    "$addFields": {
        "_vqbTempMonth" "$switch": {
            "branches": [
                {
                    "case": {"$in": month_names},
                    "then": month_number,
                }
                for month_number, month_names in MONTH_NUMBER_TO_NAMES.items()
            ],
            "default": None,
        }
    }
}


def _extract_date_parts_to_temp_fields(
    year_position: int, month_position: int | None = None, day_position: int | None = None
) -> MongoStep:
    date_parts_temp_fields: MongoStep = {
        "_vqbTempYear": {"$arrayElemAt": ["$_vqbTempArray", year_position]},
    }

    if month_position is not None:
        date_parts_temp_fields["_vqbTempMonth"] = {"$toLower": {"$arrayElemAt": ["$_vqbTempArray", month_position]}}

    if day_position is not None:
        date_parts_temp_fields["_vqbTempDay"] = {"$arrayElemAt": ["$_vqbTempArray", day_position]}

    return {
        "$addFields": date_parts_temp_fields,
    }


def _clean_temp_fields():
    return {"$project": {"_vqbTempArray": 0, "_vqbTempMonth": 0, "_vqbTempYear": 0, "_vqbTempDate": 0}}


def _concat_fields_to_date(target_col: str, fields: list[str | dict], format: str):
    return {
        "$addFields": {
            target_col: {
                "$dateFromString": {
                    "dateString": {
                        "$concat": fields,
                    },
                    "format": format,
                },
            },
        },
    }



import geopandas as gpd
import pandas as pd


class UnsupportedGeoOperation(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(f"Unsupported Geo operation: {msg}")


def df_to_geodf(df: pd.DataFrame) -> gpd.GeoDataFrame:
    if not (isinstance(df, pd.DataFrame) and hasattr(df, "geometry")):
        raise UnsupportedGeoOperation("df must be a DataFrame and 'geometry' must be set")

    # Checking this second because we can have a GeoDataFrame without geometry
    if isinstance(df, gpd.GeoDataFrame):
        return df

    try:
        return gpd.GeoDataFrame(df)
    except Exception as exc:
        raise UnsupportedGeoOperation(f"Could not convert DataFrame to GeoDataFrame: {exc}") from exc



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import CumSumStep


def execute_cumsum(
    step: CumSumStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    df = df.sort_values(step.reference_column)
    df_grouped = df.groupby(step.groupby, dropna=False) if step.groupby else df

    for col in step.to_cumsum:
        dst_column = col[1] or f"{col[0]}_CUMSUM"
        cumsum_serie = df_grouped[col[0]].cumsum()
        df = df.assign(**{dst_column: cumsum_serie})

    return df



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import MovingAverageStep


def execute_moving_average(
    step: MovingAverageStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column_name = step.new_column_name or f"{step.value_column}_MOVING_AVG"
    df = df.sort_values(by=step.groups + [step.column_to_sort]).reset_index(drop=True)
    if step.groups:
        df_grouped = df.groupby(step.groups, dropna=False)
    else:
        df_grouped = df
    serie = df_grouped.rolling(step.moving_window).mean()[step.value_column].reset_index(drop=True)
    return df.assign(**{new_column_name: serie})



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.moving_average import MovingAverageStep


def translate_moving_average(step: MovingAverageStep) -> list[MongoStep]:
    return [
        # Ensure the reference column is sorted to prepare for the moving average computation
        {"$sort": {step.column_to_sort: 1}},
        # If needing to group the computation, provide for the columns, else _id is set to None
        {
            "$group": {
                "_id": {col: f"${col}" for col in step.groups} if step.groups else None,
                "_vqbArray": {"$push": "$$ROOT"},
            },
        },
        # Prepare an array of documents with a new moving average field. One array per group (if any)
        {
            "$addFields": {
                "_vqbArray": {
                    # We use $map to apply operations while looping over the _vqbArray documents
                    "$map": {
                        # We create an index ("idx") variable that will go from 0 to the size of _vqbArray
                        "input": {"$range": [0, {"$size": "$_vqbArray"}]},
                        "as": "idx",
                        # We will use the idx variable in the following stages
                        "in": {
                            "$cond": [
                                # If the index is less than the moving window minus 1...
                                {"$lt": ["$$idx", (step.moving_window) - 1]},  # explicit type for typescript
                                # ... then we cannot apply the moving average computation, and
                                # we just keep the original document without any new field...
                                {"$arrayElemAt": ["$_vqbArray", "$$idx"]},
                                # ... else we compute the value average over the last N documents starting
                                # from the current index document, N being equal to the moving window)
                                {
                                    "$mergeObjects": [
                                        # Keep track of original document
                                        {"$arrayElemAt": ["$_vqbArray", "$$idx"]},
                                        # and add the new moving average column
                                        {
                                            step.new_column_name or f"{step.value_column}_MOVING_AVG": {
                                                "$avg": {
                                                    "$slice": [
                                                        f"$_vqbArray.{step.value_column}",
                                                        {
                                                            "$subtract": [
                                                                "$$idx",
                                                                step.moving_window - 1,
                                                            ]
                                                        },  # explicit type for typescript
                                                        step.moving_window,
                                                    ],
                                                },
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                    },
                },
            },
        },
        # Flatten the array(s) to get back to 1 row per document (at this stage in the field "_vqbArray")
        {"$unwind": "$_vqbArray"},
        # Set the _vqbArray field as the new root to get back to the original document granularity
        {"$replaceRoot": {"newRoot": "$_vqbArray"}},
    ]



import pandas as pd

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.dissolve import DissolveStep
from weaverbird.pipeline.steps.hierarchy import HierarchyStep

from .dissolve import execute_dissolve


def _dissolve_one_level(df: pd.DataFrame, groups: list[str], level_col: str, level: int) -> pd.DataFrame:
    dissolved = execute_dissolve(DissolveStep(groups=groups), df)
    dissolved[level_col] = level
    return dissolved


def execute_hierarchy(
    step: HierarchyStep,
    df: pd.DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> pd.DataFrame:
    df[step.hierarchy_level_column] = len(step.hierarchy)
    return pd.concat(
        [df]
        + [
            _dissolve_one_level(
                df,
                step.hierarchy[: -idx if idx else None],
                step.hierarchy_level_column,
                len(step.hierarchy) - idx - 1,
            )
            for idx in range(len(step.hierarchy))
        ],
        ignore_index=True,
    )



from collections.abc import Callable
from typing import Any, Protocol

from pandas import DataFrame
from pydantic import BaseModel, Field

from weaverbird.pipeline import Pipeline
from weaverbird.pipeline.steps.utils.combination import Reference


class StepExecutionReport(BaseModel):
    step_index: int
    time_spent_in_ms: int
    memory_used_in_bytes: int


class PipelineExecutionReport(BaseModel):
    steps_reports: list[StepExecutionReport] = Field(min_length=0)


DomainRetriever = Callable[[str | Reference], DataFrame]
PipelineExecutor = Callable[[Pipeline, DomainRetriever], tuple[DataFrame, PipelineExecutionReport]]


class StepExecutor(Protocol):
    def __call__(
        self,
        step: Any,
        df: DataFrame,
        domain_retriever: DomainRetriever | None,
        execute_pipeline: PipelineExecutor | None,
    ) -> DataFrame: ...



from datetime import UTC, datetime

from numpy.ma import logical_and, logical_or
from pandas import DataFrame, Series, Timestamp
from pandas import to_datetime as pd_to_datetime
from pandas.tseries.offsets import DateOffset

from weaverbird.backends.pandas_executor.steps.utils.dates import evaluate_relative_date
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.dates import RelativeDate


def _date_bound_condition_to_tz_aware_timestamp(value: str | RelativeDate | datetime) -> Timestamp:
    if isinstance(value, RelativeDate):
        value = evaluate_relative_date(value)
    if isinstance(value, datetime):
        tz = value.tzinfo or UTC
        # Cannot pass a tz-aware datetime object with tz arg
        return Timestamp(value.replace(tzinfo=None), tz=tz)
    else:  # str
        ts = Timestamp(value)
        return ts if ts.tzinfo else ts.replace(tzinfo=UTC)


def apply_condition(condition: Condition, df: DataFrame) -> Series:
    """
    Returns a boolean Series, that will be used to filter a DataFrame.

    Example:
        filter = apply_condition(condition, df)
        df[filter]  # The filtered DataFrame
    """
    if isinstance(condition, ComparisonCondition):
        return getattr(df[condition.column], condition.operator)(condition.value)
    elif isinstance(condition, InclusionCondition):
        f = df[condition.column].isin(condition.value)
        if condition.operator.startswith("n"):
            return ~f
        else:
            return f
    elif isinstance(condition, NullCondition):
        f = df[condition.column].isnull()
        if condition.operator.startswith("not"):
            return ~f
        else:
            return f
    elif isinstance(condition, MatchCondition):
        # We must ensure that the column does not contain pd.NA, as we must return only boolean
        # values
        f = df[condition.column].notnull() & df[condition.column].str.contains(condition.value)
        if condition.operator.startswith("not"):
            return ~f
        else:
            return f

    elif isinstance(condition, DateBoundCondition):
        if condition.operator == "until":
            comparison_method = "le"
        elif condition.operator == "from":
            comparison_method = "ge"
        else:
            raise NotImplementedError

        value = _date_bound_condition_to_tz_aware_timestamp(condition.value)

        # Remove time info from the column to filter on. Using utc=True to ensure we have time-aware
        # objects. Since we're only using this for comparison, the input column is not modified,
        column_without_time = pd_to_datetime(df[condition.column], utc=True) - DateOffset(
            hour=0, minute=0, second=0, microsecond=0, nanosecond=0
        )
        # Do the same with the value to compare it to
        value_without_time = value - DateOffset(hour=0, minute=0, second=0, microsecond=0, nanosecond=0)

        return getattr(column_without_time, comparison_method)(value_without_time)

    elif isinstance(condition, ConditionComboAnd):
        return logical_and.reduce([apply_condition(c, df) for c in condition.and_])
    elif isinstance(condition, ConditionComboOr):
        return logical_or.reduce([apply_condition(c, df) for c in condition.or_])
    else:
        raise NotImplementedError



from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict


def to_camelcase(string: str) -> str:
    first_word, *other_words = string.split("_")
    return first_word + "".join(word.capitalize() for word in other_words)


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="forbid", alias_generator=to_camelcase)


class BaseStep(BaseModel):
    name: str

    def model_dump(self, *, exclude_none: bool = True, **kwargs) -> dict[str, Any]:
        return super().model_dump(exclude_none=exclude_none, **kwargs)

    # None values are excluded, to avoid triggering validations error in front-ends
    def dict(self, *, exclude_none: bool = True, **kwargs) -> dict[str, Any]:
        return self.model_dump(exclude_none=exclude_none, **kwargs)



import itertools
from typing import Any


def combinations(items: list[Any]) -> list[tuple[Any, ...]]:
    # combinations of all lenghts must be included. return is ordered by combination lenght
    result = []
    for i in range(1, len(items) + 1):
        result += list(itertools.combinations(items, i))
    return result
